"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _objectSpread2 = _interopRequireDefault(require("@babel/runtime/helpers/objectSpread"));

var _path = _interopRequireDefault(require("path"));

var _worker_threads = require("worker_threads");

var _commands = _interopRequireDefault(require("./worker/commands"));

var _defaults = _interopRequireDefault(require("./worker/defaults"));

var _atomics = _interopRequireDefault(require("./worker/atomics"));

// Copyright 2017-2018 @polkadot/client-db authors & contributors
// This software may be modified and distributed under the terms
// of the ISC license. See the LICENSE file for details.
const emptyBuffer = new Uint8Array();

class SyncDb {
  constructor(type = 'memory', dbPath = '.', isTrie = true) {
    this.worker = void 0;
    // NOTE Node 10.6 relative paths for Workers are broken - adding here tries to load
    // the worker from /client, not client-db.
    this.worker = new _worker_threads.Worker(_path.default.join(__dirname, './worker/index.js'), {
      workerData: {
        isTrie,
        path: dbPath,
        type
      }
    });
  }

  checkpoint() {
    this._executeWrite('checkpoint');
  }

  commit() {
    this._executeWrite('commit');
  }

  revert() {
    this._executeWrite('revert');
  }

  del(key) {
    this._executeWrite('del', key);
  }

  get(key) {
    return this._executeRead('get', key);
  }

  put(key, value) {
    this._executeWrite('put', key, value);
  }

  getRoot() {
    return this._executeRead('getRoot');
  }

  setRoot(value) {
    this._executeWrite('setRoot', undefined, value);
  }

  async terminate() {
    this.worker.unref(); // TODO We should cleanup the trie instance gracefully, so another message here
    // to cleanup and then the termination up next
    // TODO We should be terminating the worker when no references
    // return promisify(this.worker, this.worker.terminate);
  } // Sends a message to the worker, waiting until started


  _waitOnStart(type, message) {
    const state = new Int32Array(new SharedArrayBuffer(8));
    this.worker.postMessage((0, _objectSpread2.default)({}, message, {
      state,
      type
    }));

    _atomics.default.wait(state, _commands.default.START);

    return state;
  } // Notifies the worker that it should continue filling the result buffer


  _waitOnRead(state) {
    _atomics.default.notify(state, _commands.default.FILL);
  } // Ok, this is not something that returns a value, just send the message and
  // return when we the call has been done


  _executeWrite(type, key, value) {
    this._waitOnStart(type, {
      buffer: emptyBuffer,
      key,
      value
    });
  } // Sends a message to the worker, reading and returning the actual result


  _executeRead(type, key, value) {
    // The shared data buffer that will be used by the worker to send info back
    const shared = new SharedArrayBuffer(_defaults.default.SHARED_BUFFER_SIZE);
    const buffer = new Uint8Array(shared);

    const state = this._waitOnStart(type, {
      buffer,
      key,
      value
    });

    const size = this._readSize(state, shared);

    if (size === -1) {
      return null;
    }

    return this._readBuffer(state, buffer, size);
  } // Read the size of a structure to be returned from the stream


  _readSize(state, shared) {
    const view = new DataView(shared); // expect to read SIZE, END/ERROR here

    switch (state[0]) {
      case _commands.default.END:
      case _commands.default.ERROR:
        return -1;
      // Read the size to detemine how big of a result buffer we need.

      case _commands.default.SIZE:
        return view.getUint32(0);
      // This _should_ never happen... but...

      default:
        throw new Error(`Unknown worker state waiting for size, command=${state[0]}`);
    }
  } // Read the full data buffer from the worker


  _readBuffer(state, buffer, size) {
    const result = new Uint8Array(size);
    let available = 0;
    let written = 0; // Here we loop through the states and either read data to fill the buffer
    // or return when it is time to do so

    while (written !== size) {
      this._waitOnRead(state);

      switch (state[0]) {
        // Error, so just return a null for the caller to handle
        case _commands.default.ERROR:
          return null;
        // Get the available data from the buffer and write it into our result

        case _commands.default.END:
        case _commands.default.READ:
          available = Math.min(buffer.length, size - written);
          result.set(buffer.subarray(0, available), written);
          written += available;
          break;
        // This _should_ never happen... but...

        default:
          throw new Error(`Unknown worker state waiting for read, ${state[0]}`);
      }
    }

    return result;
  }

}

exports.default = SyncDb;